@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: ogr_spatialref.h 28972 2015-04-22 10:39:11Z rouault $
+ * $Id: ogr_spatialref.h e37e476c4cf8f4b0df8995e0d95d5d672fca1a9b 2018-05-05 16:54:18 +0200 Even Rouault $
  *
  * Project:  OpenGIS Simple Features Reference Implementation
  * Purpose:  Classes for manipulating spatial reference systems in a
@@ -29,11 +29,16 @@
  * DEALINGS IN THE SOFTWARE.
  ****************************************************************************/
 
-#ifndef _OGR_SPATIALREF_H_INCLUDED
-#define _OGR_SPATIALREF_H_INCLUDED
+#ifndef OGR_SPATIALREF_H_INCLUDED
+#define OGR_SPATIALREF_H_INCLUDED
 
+#include "cpl_string.h"
 #include "ogr_srs_api.h"
 
+#include <cstddef>
+#include <map>
+#include <vector>
+
 /**
  * \file ogr_spatialref.h
  *
@@ -52,8 +57,8 @@
  * second with a value of 1.
  *
  * Normally application code just interacts with the OGRSpatialReference
- * object, which uses the OGR_SRSNode to implement it's data structure;
- * however, this class is user accessable for detailed access to components
+ * object, which uses the OGR_SRSNode to implement its data structure;
+ * however, this class is user accessible for detailed access to components
  * of an SRS definition.
  */
 
@@ -67,14 +72,19 @@ class CPL_DLL OGR_SRSNode
     int         nChildren;
 
     int         NeedsQuoting() const;
-    OGRErr      importFromWkt( char **, int nRecLevel, int* pnNodes );
-    
+    OGRErr      importFromWkt( const char **, int nRecLevel, int* pnNodes );
+
+    CPL_DISALLOW_COPY_ASSIGN(OGR_SRSNode)
+
   public:
-                OGR_SRSNode(const char * = NULL);
+    explicit     OGR_SRSNode(const char * = nullptr);
                 ~OGR_SRSNode();
 
+    /** Return whether this is a leaf node.
+     * @return TRUE or FALSE
+     */
     int         IsLeafNode() const { return nChildren == 0; }
-    
+
     int         GetChildCount() const { return nChildren; }
     OGR_SRSNode *GetChild( int );
     const OGR_SRSNode *GetChild( int ) const;
@@ -97,13 +107,18 @@ class CPL_DLL OGR_SRSNode
 
     OGR_SRSNode *Clone() const;
 
-    OGRErr      importFromWkt( char ** );
+    OGRErr      importFromWkt( char ** )
+/*! @cond Doxygen_Suppress */
+                    CPL_WARN_DEPRECATED("Use importFromWkt(const char**)")
+/*! @endcond */
+                    ;
+    OGRErr      importFromWkt( const char ** );
     OGRErr      exportToWkt( char ** ) const;
     OGRErr      exportToPrettyWkt( char **, int = 1) const;
-    
-    OGRErr      applyRemapper( const char *pszNode, 
-                               char **papszSrcValues, 
-                               char **papszDstValues, 
+
+    OGRErr      applyRemapper( const char *pszNode,
+                               const char * const *papszSrcValues,
+                               const char * const *papszDstValues,
                                int nStepSize = 1,
                                int bChildOfHit = FALSE );
 };
@@ -113,7 +128,7 @@ class CPL_DLL OGR_SRSNode
 /************************************************************************/
 
 /**
- * This class respresents a OpenGIS Spatial Reference System, and contains
+ * This class represents an OpenGIS Spatial Reference System, and contains
  * methods for converting between this object organization and well known
  * text (WKT) format.  This object is reference counted as one instance of
  * the object is normally shared between many OGRGeometry objects.
@@ -124,18 +139,21 @@ class CPL_DLL OGR_SRSNode
  *
  * See <a href="osr_tutorial.html">the tutorial</a> for more information on
  * how to use this class.
+ * 
+ * Consult also the <a href="wktproblems.html">OGC WKT Coordinate System Issues</a> page
+ * for implementation details of WKT in OGR.
  */
 
 class CPL_DLL OGRSpatialReference
 {
-    double      dfFromGreenwich;
-    double      dfToMeter;
-    double      dfToDegrees;
+    mutable double      dfFromGreenwich;
+    mutable double      dfToMeter;
+    mutable double      dfToDegrees;
 
     OGR_SRSNode *poRoot;
 
     int         nRefCount;
-    int         bNormInfoSet;
+    mutable int         bNormInfoSet;
 
     static OGRErr Validate(OGR_SRSNode *poRoot);
     static OGRErr ValidateAuthority(OGR_SRSNode *poRoot);
@@ -149,14 +167,22 @@ class CPL_DLL OGRSpatialReference
     OGRErr      importFromURNPart(const char* pszAuthority,
                                   const char* pszCode,
                                   const char* pszURN);
+
+    OGRErr      importFromEPSGAInternal(int nCode,
+                                        const char* pszSRSType);
+
+    static const std::vector<OGRSpatialReference*>* GetSRSCache(
+                    const char* pszSRSType,
+                    const std::map<CPLString, int>*& poMapESRICSNameToCodeOut);
+
   public:
                 OGRSpatialReference(const OGRSpatialReference&);
-                OGRSpatialReference(const char * = NULL);
-                
+    explicit    OGRSpatialReference(const char * = nullptr);
+
     virtual    ~OGRSpatialReference();
-    
+
     static void DestroySpatialReference(OGRSpatialReference* poSRS);
-                
+
     OGRSpatialReference &operator=(const OGRSpatialReference&);
 
     int         Reference();
@@ -173,24 +199,33 @@ class CPL_DLL OGRSpatialReference
     OGRErr      exportToProj4( char ** ) const;
     OGRErr      exportToPCI( char **, char **, double ** ) const;
     OGRErr      exportToUSGS( long *, long *, double **, long * ) const;
-    OGRErr      exportToXML( char **, const char * = NULL ) const;
+    OGRErr      exportToXML( char **, const char * = nullptr ) const;
     OGRErr      exportToPanorama( long *, long *, long *, long *,
                                   double * ) const;
     OGRErr      exportToERM( char *pszProj, char *pszDatum, char *pszUnits );
     OGRErr      exportToMICoordSys( char ** ) const;
-    
-    OGRErr      importFromWkt( char ** );
+
+
+    OGRErr      importFromWkt( char ** )
+/*! @cond Doxygen_Suppress */
+        CPL_WARN_DEPRECATED("Use importFromWkt(const char**) or importFromWkt(const char*)")
+/*! @endcond */
+        ;
+
+    OGRErr      importFromWkt( const char ** );
+    OGRErr      importFromWkt( const char* );
     OGRErr      importFromProj4( const char * );
     OGRErr      importFromEPSG( int );
     OGRErr      importFromEPSGA( int );
     OGRErr      importFromESRI( char ** );
-    OGRErr      importFromPCI( const char *, const char * = NULL,
-                               double * = NULL );
-#define USGS_ANGLE_DECIMALDEGREES 0
-#define USGS_ANGLE_PACKEDDMS      TRUE /* 1 */
-#define USGS_ANGLE_RADIANS        2
+    OGRErr      importFromPCI( const char *, const char * = nullptr,
+                               double * = nullptr );
+
+#define USGS_ANGLE_DECIMALDEGREES 0     /**< Angle is in decimal degrees. */
+#define USGS_ANGLE_PACKEDDMS      TRUE  /**< Angle is in packed degree minute second. */
+#define USGS_ANGLE_RADIANS        2     /**< Angle is in radians. */
     OGRErr      importFromUSGS( long iProjSys, long iZone,
-                                double *padfPrjParams, long iDatum, 
+                                double *padfPrjParams, long iDatum,
                                 int nUSGSAngleFormat = USGS_ANGLE_PACKEDDMS );
     OGRErr      importFromPanorama( long, long, long, double* );
     OGRErr      importFromOzi( const char * const* papszLines );
@@ -203,31 +238,38 @@ class CPL_DLL OGRSpatialReference
                                const char *pszUnits );
     OGRErr      importFromUrl( const char * );
     OGRErr      importFromMICoordSys( const char * );
-    
+
     OGRErr      morphToESRI();
     OGRErr      morphFromESRI();
 
-    OGRErr      Validate();
-    OGRErr      StripCTParms( OGR_SRSNode * = NULL );
+    OGRSpatialReference* convertToOtherProjection(
+                                    const char* pszTargetProjection,
+                                    const char* const* papszOptions = nullptr ) const;
+
+    OGRErr      Validate() const;
+    OGRErr      StripCTParms( OGR_SRSNode * = nullptr );
     OGRErr      StripVertical();
     OGRErr      FixupOrdering();
     OGRErr      Fixup();
 
-    int         EPSGTreatsAsLatLong();
-    int         EPSGTreatsAsNorthingEasting();
-    const char *GetAxis( const char *pszTargetKey, int iAxis, 
+    int         EPSGTreatsAsLatLong() const;
+    int         EPSGTreatsAsNorthingEasting() const;
+    const char *GetAxis( const char *pszTargetKey, int iAxis,
                          OGRAxisOrientation *peOrientation ) const;
-    OGRErr      SetAxes( const char *pszTargetKey, 
-                         const char *pszXAxisName, 
+    OGRErr      SetAxes( const char *pszTargetKey,
+                         const char *pszXAxisName,
                          OGRAxisOrientation eXAxisOrientation,
-                         const char *pszYAxisName, 
+                         const char *pszYAxisName,
                          OGRAxisOrientation eYAxisOrientation );
 
-    // Machinary for accessing parse nodes
+    // Machinery for accessing parse nodes
+
+    //! Return root node
     OGR_SRSNode *GetRoot() { return poRoot; }
+    //! Return root node
     const OGR_SRSNode *GetRoot() const { return poRoot; }
     void        SetRoot( OGR_SRSNode * );
-    
+
     OGR_SRSNode *GetAttrNode(const char *);
     const OGR_SRSNode *GetAttrNode(const char *) const;
     const char  *GetAttrValue(const char *, int = 0) const;
@@ -235,19 +277,43 @@ class CPL_DLL OGRSpatialReference
     OGRErr      SetNode( const char *, const char * );
     OGRErr      SetNode( const char *, double );
 
-    OGRErr      SetLinearUnitsAndUpdateParameters( const char *pszName, 
+    OGRErr      SetLinearUnitsAndUpdateParameters( const char *pszName,
                                                    double dfInMeters );
     OGRErr      SetLinearUnits( const char *pszName, double dfInMeters );
     OGRErr      SetTargetLinearUnits( const char *pszTargetKey,
                                       const char *pszName, double dfInMeters );
-    double      GetLinearUnits( char ** = NULL ) const;
+
+    double      GetLinearUnits( char ** ) const CPL_WARN_DEPRECATED("Use GetLinearUnits(const char**) instead");
+    double      GetLinearUnits( const char ** = nullptr ) const;
+/*! @cond Doxygen_Suppress */
+    double      GetLinearUnits( std::nullptr_t ) const
+        { return GetLinearUnits( static_cast<const char**>(nullptr) ); }
+/*! @endcond */
+
     double      GetTargetLinearUnits( const char *pszTargetKey,
-                                      char ** ppszRetName = NULL ) const;
+                                      char ** ppszRetName ) const
+            CPL_WARN_DEPRECATED("Use GetTargetLinearUnits(const char*, const char**)");
+    double      GetTargetLinearUnits( const char *pszTargetKey,
+                                      const char ** ppszRetName = nullptr ) const;
+/*! @cond Doxygen_Suppress */
+    double      GetTargetLinearUnits( const char *pszTargetKey, std::nullptr_t ) const
+        { return GetTargetLinearUnits( pszTargetKey, static_cast<const char**>(nullptr) ); }
+/*! @endcond */
 
     OGRErr      SetAngularUnits( const char *pszName, double dfInRadians );
-    double      GetAngularUnits( char ** = NULL ) const;
+    double      GetAngularUnits( char ** ) const CPL_WARN_DEPRECATED("Use GetAngularUnits(const char**) instead");
+    double      GetAngularUnits( const char ** = nullptr ) const;
+/*! @cond Doxygen_Suppress */
+    double      GetAngularUnits( std::nullptr_t ) const
+        { return GetAngularUnits( static_cast<const char**>(nullptr) ); }
+/*! @endcond */
 
-    double      GetPrimeMeridian( char ** = NULL ) const;
+    double      GetPrimeMeridian( char ** ) const CPL_WARN_DEPRECATED("Use GetPrimeMeridian(const char**) instead");
+    double      GetPrimeMeridian( const char ** = nullptr ) const;
+/*! @cond Doxygen_Suppress */
+    double      GetPrimeMeridian( std::nullptr_t ) const
+        { return GetPrimeMeridian( static_cast<const char**>(nullptr) ); }
+/*! @endcond */
 
     int         IsGeographic() const;
     int         IsProjected() const;
@@ -256,8 +322,12 @@ class CPL_DLL OGRSpatialReference
     int         IsVertical() const;
     int         IsCompound() const;
     int         IsSameGeogCS( const OGRSpatialReference * ) const;
+    int         IsSameGeogCS( const OGRSpatialReference *,
+                              const char* const * papszOptions ) const;
     int         IsSameVertCS( const OGRSpatialReference * ) const;
     int         IsSame( const OGRSpatialReference * ) const;
+    int         IsSame( const OGRSpatialReference *,
+                        const char* const * papszOptions ) const;
 
     void        Clear();
     OGRErr      SetLocalCS( const char * );
@@ -268,16 +338,16 @@ class CPL_DLL OGRSpatialReference
                            const char * pszDatumName,
                            const char * pszEllipsoidName,
                            double dfSemiMajor, double dfInvFlattening,
-                           const char * pszPMName = NULL,
+                           const char * pszPMName = nullptr,
                            double dfPMOffset = 0.0,
-                           const char * pszUnits = NULL,
+                           const char * pszUnits = nullptr,
                            double dfConvertToRadians = 0.0 );
     OGRErr      SetWellKnownGeogCS( const char * );
     OGRErr      CopyGeogCSFrom( const OGRSpatialReference * poSrcSRS );
     OGRErr      SetVertCS( const char *pszVertCSName,
-                           const char *pszVertDatumName, 
+                           const char *pszVertDatumName,
                            int nVertDatumClass = 2005 );
-    OGRErr      SetCompoundCS( const char *pszName, 
+    OGRErr      SetCompoundCS( const char *pszName,
                                const OGRSpatialReference *poHorizSRS,
                                const OGRSpatialReference *poVertSRS );
 
@@ -287,35 +357,41 @@ class CPL_DLL OGRSpatialReference
                             double = 0.0, double = 0.0, double = 0.0,
                             double = 0.0 );
     OGRErr      GetTOWGS84( double *padfCoef, int nCoeff = 7 ) const;
-    
-    double      GetSemiMajor( OGRErr * = NULL ) const;
-    double      GetSemiMinor( OGRErr * = NULL ) const;
-    double      GetInvFlattening( OGRErr * = NULL ) const;
 
-    OGRErr      SetAuthority( const char * pszTargetKey, 
-                              const char * pszAuthority, 
+    double      GetSemiMajor( OGRErr * = nullptr ) const;
+    double      GetSemiMinor( OGRErr * = nullptr ) const;
+    double      GetInvFlattening( OGRErr * = nullptr ) const;
+    double      GetEccentricity() const;
+    double      GetSquaredEccentricity() const;
+
+    OGRErr      SetAuthority( const char * pszTargetKey,
+                              const char * pszAuthority,
                               int nCode );
 
     OGRErr      AutoIdentifyEPSG();
-    int         GetEPSGGeogCS();
+    OGRSpatialReferenceH* FindMatches( char** papszOptions,
+                                       int* pnEntries,
+                                       int** ppanMatchConfidence ) const;
+
+    int         GetEPSGGeogCS() const;
 
     const char *GetAuthorityCode( const char * pszTargetKey ) const;
     const char *GetAuthorityName( const char * pszTargetKey ) const;
 
-    const char *GetExtension( const char *pszTargetKey, 
+    const char *GetExtension( const char *pszTargetKey,
+                              const char *pszName,
+                              const char *pszDefault = nullptr ) const;
+    OGRErr      SetExtension( const char *pszTargetKey,
                               const char *pszName,
-                              const char *pszDefault = NULL ) const;
-    OGRErr      SetExtension( const char *pszTargetKey, 
-                              const char *pszName, 
                               const char *pszValue );
-    
+
     int         FindProjParm( const char *pszParameter,
-                              const OGR_SRSNode *poPROJCS=NULL ) const;
+                              const OGR_SRSNode *poPROJCS=nullptr ) const;
     OGRErr      SetProjParm( const char *, double );
-    double      GetProjParm( const char *, double =0.0, OGRErr* = NULL ) const;
+    double      GetProjParm( const char *, double =0.0, OGRErr* = nullptr ) const;
 
     OGRErr      SetNormProjParm( const char *, double );
-    double      GetNormProjParm( const char *, double=0.0, OGRErr* =NULL)const;
+    double      GetNormProjParm( const char *, double=0.0, OGRErr* =nullptr)const;
 
     static int  IsAngularParameter( const char * );
     static int  IsLongitudeParameter( const char * );
@@ -325,7 +401,7 @@ class CPL_DLL OGRSpatialReference
     OGRErr      SetACEA( double dfStdP1, double dfStdP2,
                          double dfCenterLat, double dfCenterLong,
                          double dfFalseEasting, double dfFalseNorthing );
-    
+
     /** Azimuthal Equidistant */
     OGRErr      SetAE( double dfCenterLat, double dfCenterLong,
                        double dfFalseEasting, double dfFalseNorthing );
@@ -333,7 +409,7 @@ class CPL_DLL OGRSpatialReference
     /** Bonne */
     OGRErr      SetBonne( double dfStdP1, double dfCentralMeridian,
                           double dfFalseEasting, double dfFalseNorthing );
-    
+
     /** Cylindrical Equal Area */
     OGRErr      SetCEA( double dfStdP1, double dfCentralMeridian,
                         double dfFalseEasting, double dfFalseNorthing );
@@ -347,13 +423,15 @@ class CPL_DLL OGRSpatialReference
                        double dfCenterLat, double dfCenterLong,
                        double dfFalseEasting, double dfFalseNorthing );
 
-    /** Eckert I-VI */
+    /** Eckert I */
     OGRErr      SetEckert( int nVariation, double dfCentralMeridian,
                            double dfFalseEasting, double dfFalseNorthing );
 
+    /** Eckert IV */
     OGRErr      SetEckertIV( double dfCentralMeridian,
                              double dfFalseEasting, double dfFalseNorthing );
 
+    /** Eckert VI */
     OGRErr      SetEckertVI( double dfCentralMeridian,
                              double dfFalseEasting, double dfFalseNorthing );
 
@@ -366,11 +444,11 @@ class CPL_DLL OGRSpatialReference
                                      double dfFalseEasting, double dfFalseNorthing );
 
     /** Geostationary Satellite */
-    OGRErr      SetGEOS( double dfCentralMeridian, double dfSatelliteHeight, 
+    OGRErr      SetGEOS( double dfCentralMeridian, double dfSatelliteHeight,
                          double dfFalseEasting, double dfFalseNorthing );
 
     /** Goode Homolosine */
-    OGRErr      SetGH( double dfCentralMeridian, 
+    OGRErr      SetGH( double dfCentralMeridian,
                        double dfFalseEasting, double dfFalseNorthing );
 
     /** Interrupted Goode Homolosine */
@@ -379,7 +457,7 @@ class CPL_DLL OGRSpatialReference
     /** Gall Stereograpic */
     OGRErr      SetGS( double dfCentralMeridian,
                        double dfFalseEasting, double dfFalseNorthing );
- 
+
     /** Gauss Schreiber Transverse Mercator */
     OGRErr      SetGaussSchreiberTMercator(double dfCenterLat, double dfCenterLong,
                                            double dfScale,
@@ -395,16 +473,20 @@ class CPL_DLL OGRSpatialReference
                         double dfScale,
                         double dfFalseEasting, double dfFalseNorthing );
 
+    /**  Hotine Oblique Mercator 2 points */
     OGRErr      SetHOM2PNO( double dfCenterLat,
                             double dfLat1, double dfLong1,
                             double dfLat2, double dfLong2,
                             double dfScale,
                             double dfFalseEasting, double dfFalseNorthing );
 
+#ifdef undef
+    /** Oblique Mercator */
     OGRErr      SetOM( double dfCenterLat, double dfCenterLong,
                        double dfAzimuth, double dfRectToSkew,
                        double dfScale,
                        double dfFalseEasting, double dfFalseNorthing );
+#endif
 
     /** Hotine Oblique Mercator Azimuth Center / Variant B */
     OGRErr      SetHOMAC( double dfCenterLat, double dfCenterLong,
@@ -421,7 +503,7 @@ class CPL_DLL OGRSpatialReference
     /** Krovak Oblique Conic Conformal */
     OGRErr      SetKrovak( double dfCenterLat, double dfCenterLong,
                            double dfAzimuth, double dfPseudoStdParallelLat,
-                           double dfScale, 
+                           double dfScale,
                            double dfFalseEasting, double dfFalseNorthing );
 
     /** Lambert Azimuthal Equal-Area */
@@ -442,16 +524,17 @@ class CPL_DLL OGRSpatialReference
     OGRErr      SetLCCB( double dfStdP1, double dfStdP2,
                          double dfCenterLat, double dfCenterLong,
                          double dfFalseEasting, double dfFalseNorthing );
-    
+
     /** Miller Cylindrical */
     OGRErr      SetMC( double dfCenterLat, double dfCenterLong,
                        double dfFalseEasting, double dfFalseNorthing );
 
-    /** Mercator */
+    /** Mercator 1SP */
     OGRErr      SetMercator( double dfCenterLat, double dfCenterLong,
-                             double dfScale, 
+                             double dfScale,
                              double dfFalseEasting, double dfFalseNorthing );
 
+    /** Mercator 2SP */
     OGRErr      SetMercator2SP( double dfStdP1,
                                 double dfCenterLat, double dfCenterLong,
                                 double dfFalseEasting, double dfFalseNorthing );
@@ -468,7 +551,7 @@ class CPL_DLL OGRSpatialReference
     OGRErr      SetOS( double dfOriginLat, double dfCMeridian,
                        double dfScale,
                        double dfFalseEasting,double dfFalseNorthing);
-    
+
     /** Orthographic */
     OGRErr      SetOrthographic( double dfCenterLat, double dfCenterLong,
                                  double dfFalseEasting,double dfFalseNorthing);
@@ -481,15 +564,15 @@ class CPL_DLL OGRSpatialReference
     OGRErr      SetPS( double dfCenterLat, double dfCenterLong,
                        double dfScale,
                        double dfFalseEasting, double dfFalseNorthing);
-    
+
     /** Robinson */
-    OGRErr      SetRobinson( double dfCenterLong, 
+    OGRErr      SetRobinson( double dfCenterLong,
                              double dfFalseEasting, double dfFalseNorthing );
-    
+
     /** Sinusoidal */
-    OGRErr      SetSinusoidal( double dfCenterLong, 
+    OGRErr      SetSinusoidal( double dfCenterLong,
                                double dfFalseEasting, double dfFalseNorthing );
-    
+
     /** Stereographic */
     OGRErr      SetStereographic( double dfCenterLat, double dfCenterLong,
                                   double dfScale,
@@ -498,20 +581,20 @@ class CPL_DLL OGRSpatialReference
     /** Swiss Oblique Cylindrical */
     OGRErr      SetSOC( double dfLatitudeOfOrigin, double dfCentralMeridian,
                         double dfFalseEasting, double dfFalseNorthing );
-    
+
     /** Transverse Mercator */
     OGRErr      SetTM( double dfCenterLat, double dfCenterLong,
                        double dfScale,
                        double dfFalseEasting, double dfFalseNorthing );
 
     /** Transverse Mercator variants. */
-    OGRErr      SetTMVariant( const char *pszVariantName, 
+    OGRErr      SetTMVariant( const char *pszVariantName,
                               double dfCenterLat, double dfCenterLong,
                               double dfScale,
                               double dfFalseEasting, double dfFalseNorthing );
 
     /** Tunesia Mining Grid  */
-    OGRErr      SetTMG( double dfCenterLat, double dfCenterLong, 
+    OGRErr      SetTMG( double dfCenterLat, double dfCenterLong,
                         double dfFalseEasting, double dfFalseNorthing );
 
     /** Transverse Mercator (South Oriented) */
@@ -520,17 +603,17 @@ class CPL_DLL OGRSpatialReference
                          double dfFalseEasting, double dfFalseNorthing );
 
     /** Two Point Equidistant */
-    OGRErr      SetTPED( double dfLat1, double dfLong1, 
-                         double dfLat2, double dfLong2, 
+    OGRErr      SetTPED( double dfLat1, double dfLong1,
+                         double dfLat2, double dfLong2,
                          double dfFalseEasting, double dfFalseNorthing );
-    
+
     /** VanDerGrinten */
     OGRErr      SetVDG( double dfCenterLong,
                         double dfFalseEasting, double dfFalseNorthing );
 
     /** Universal Transverse Mercator */
     OGRErr      SetUTM( int nZone, int bNorth = TRUE );
-    int         GetUTMZone( int *pbNorth = NULL ) const;
+    int         GetUTMZone( int *pbNorth = nullptr ) const;
 
     /** Wagner I -- VII */
     OGRErr      SetWagner( int nVariation, double dfCenterLat,
@@ -539,19 +622,38 @@ class CPL_DLL OGRSpatialReference
     /** Quadrilateralized Spherical Cube */
     OGRErr      SetQSC(double dfCenterLat, double dfCenterLong);
 
+    /** Spherical, Cross-track, Height */
+    OGRErr      SetSCH( double dfPegLat, double dfPegLong,
+                        double dfPegHeading, double dfPegHgt);
     /** State Plane */
     OGRErr      SetStatePlane( int nZone, int bNAD83 = TRUE,
-                               const char *pszOverrideUnitName = NULL,
+                               const char *pszOverrideUnitName = nullptr,
                                double dfOverrideUnit = 0.0 );
 
-    OGRErr      ImportFromESRIStatePlaneWKT( 
-        int nCode, const char* pszDatumName, const char* pszUnitsName, 
-        int nPCSCode, const char* pszCSName = 0 );
-    OGRErr      ImportFromESRIWisconsinWKT( 
-        const char* pszPrjName, double dfCentralMeridian, double dfLatOfOrigin, 
-        const char* pszUnitsName, const char* pszCSName = 0 );
+    /** ImportFromESRIStatePlaneWKT */
+    OGRErr      ImportFromESRIStatePlaneWKT(
+        int nCode, const char* pszDatumName, const char* pszUnitsName,
+        int nPCSCode, const char* pszCSName = nullptr );
+
+    /** ImportFromESRIWisconsinWKT */
+    OGRErr      ImportFromESRIWisconsinWKT(
+        const char* pszPrjName, double dfCentralMeridian, double dfLatOfOrigin,
+        const char* pszUnitsName, const char* pszCSName = nullptr );
 
     static OGRSpatialReference* GetWGS84SRS();
+
+    /** Convert a OGRSpatialReference* to a OGRSpatialReferenceH.
+     * @since GDAL 2.3
+     */
+    static inline OGRSpatialReferenceH ToHandle(OGRSpatialReference* poSRS)
+        { return reinterpret_cast<OGRSpatialReferenceH>(poSRS); }
+
+    /** Convert a OGRSpatialReferenceH to a OGRSpatialReference*.
+     * @since GDAL 2.3
+     */
+    static inline OGRSpatialReference* FromHandle(OGRSpatialReferenceH hSRS)
+        { return reinterpret_cast<OGRSpatialReference*>(hSRS); }
+
 };
 
 /************************************************************************/
@@ -569,12 +671,12 @@ class CPL_DLL OGRSpatialReference
  *
  * Also, see OGRCreateCoordinateTransformation() for creating transformations.
  */
- 
+
 class CPL_DLL OGRCoordinateTransformation
 {
 public:
     virtual ~OGRCoordinateTransformation() {}
-    
+
     static void DestroyCT(OGRCoordinateTransformation* poCT);
 
     // From CT_CoordinateTransformation
@@ -585,6 +687,12 @@ public:
     /** Fetch internal target coordinate system. */
     virtual OGRSpatialReference *GetTargetCS() = 0;
 
+    /** Whether the transformer will emit CPLError */
+    virtual bool GetEmitErrors() { return false; }
+
+    /** Set if the transformer must emit CPLError */
+    virtual void SetEmitErrors(bool /*bEmitErrors*/) {}
+
     // From CT_MathTransform
 
     /**
@@ -592,8 +700,8 @@ public:
      *
      * This method is the same as the C function OCTTransform().
      *
-     * The method TransformEx() allows extended success information to 
-     * be captured indicating which points failed to transform. 
+     * The method TransformEx() allows extended success information to
+     * be captured indicating which points failed to transform.
      *
      * @param nCount number of points to transform.
      * @param x array of nCount X vertices, modified in place.
@@ -602,8 +710,8 @@ public:
      * @return TRUE on success, or FALSE if some or all points fail to
      * transform.
      */
-    virtual int Transform( int nCount, 
-                           double *x, double *y, double *z = NULL ) = 0;
+    virtual int Transform( int nCount,
+                           double *x, double *y, double *z = nullptr ) = 0;
 
     /**
      * Transform points from source to destination space.
@@ -614,20 +722,31 @@ public:
      * @param x array of nCount X vertices, modified in place.
      * @param y array of nCount Y vertices, modified in place.
      * @param z array of nCount Z vertices, modified in place.
-     * @param pabSuccess array of per-point flags set to TRUE if that point 
+     * @param pabSuccess array of per-point flags set to TRUE if that point
      * transforms, or FALSE if it does not.
      *
-     * @return TRUE if some or all points transform successfully, or FALSE if 
+     * @return TRUE if some or all points transform successfully, or FALSE if
      * if none transform.
      */
-    virtual int TransformEx( int nCount, 
-                             double *x, double *y, double *z = NULL,
-                             int *pabSuccess = NULL ) = 0;
+    virtual int TransformEx( int nCount,
+                             double *x, double *y, double *z = nullptr,
+                             int *pabSuccess = nullptr ) = 0;
 
+    /** Convert a OGRCoordinateTransformation* to a OGRCoordinateTransformationH.
+     * @since GDAL 2.3
+     */
+    static inline OGRCoordinateTransformationH ToHandle(OGRCoordinateTransformation* poCT)
+        { return reinterpret_cast<OGRCoordinateTransformationH>(poCT); }
+
+    /** Convert a OGRCoordinateTransformationH to a OGRCoordinateTransformation*.
+     * @since GDAL 2.3
+     */
+    static inline OGRCoordinateTransformation* FromHandle(OGRCoordinateTransformationH hCT)
+        { return reinterpret_cast<OGRCoordinateTransformation*>(hCT); }
 };
 
 OGRCoordinateTransformation CPL_DLL *
-OGRCreateCoordinateTransformation( OGRSpatialReference *poSource, 
+OGRCreateCoordinateTransformation( OGRSpatialReference *poSource,
                                    OGRSpatialReference *poTarget );
 
-#endif /* ndef _OGR_SPATIALREF_H_INCLUDED */
+#endif /* ndef OGR_SPATIALREF_H_INCLUDED */
